@@ -51,9 +51,6 @@ static ssize_t bc_give(struct bufferchain *bc, unsigned char *out, ssize_t size)
 static ssize_t bc_skip(struct bufferchain *bc, ssize_t count);
 static ssize_t bc_seekback(struct bufferchain *bc, ssize_t count);
 static void bc_forget(struct bufferchain *bc);
-#else
-#define bc_init(a)
-#define bc_reset(a)
 #endif
 
 /* A normal read and a read with timeout. */
@@ -155,6 +152,7 @@ static ssize_t icy_fullread(mpg123_handle *fr, unsigned char *buf, ssize_t count
 			{
 				/* we have got some metadata */
 				char *meta_buff;
+				/* TODO: Get rid of this malloc ... perhaps hooking into the reader buffer pool? */
 				meta_buff = malloc(meta_size+1);
 				if(meta_buff != NULL)
 				{
@@ -204,6 +202,9 @@ static ssize_t plain_fullread(mpg123_handle *fr,unsigned char *buf, ssize_t coun
 {
 	ssize_t ret,cnt=0;
 
+#ifdef EXTRA_DEBUG
+	debug1("plain fullread of %"SSIZE_P, (size_p)count);
+#endif
 	/*
 		There used to be a check for expected file end here (length value or ID3 flag).
 		This is not needed:
@@ -241,7 +242,9 @@ static void stream_close(mpg123_handle *fr)
 
 	fr->rdat.filept = 0;
 
+#ifndef NO_FEEDER
 	if(fr->rdat.flags & READER_BUFFERED)  bc_reset(&fr->rdat.buffer);
+#endif
 	if(fr->rdat.flags & READER_HANDLEIO)
 	{
 		if(fr->rdat.cleanup_handle != NULL) fr->rdat.cleanup_handle(fr->rdat.iohandle);
@@ -347,6 +350,7 @@ static off_t stream_skip_bytes(mpg123_handle *fr,off_t len)
 		}
 		return fr->rd->tell(fr);
 	}
+#ifndef NO_FEEDER
 	else if(fr->rdat.flags & READER_BUFFERED)
 	{ /* Perhaps we _can_ go a bit back. */
 		if(fr->rdat.buffer.pos >= -len)
@@ -360,6 +364,7 @@ static off_t stream_skip_bytes(mpg123_handle *fr,off_t len)
 			return READER_ERROR;
 		}
 	}
+#endif
 	else
 	{
 		fr->err = MPG123_NO_SEEK;
@@ -394,8 +399,10 @@ static int generic_read_frame_body(mpg123_handle *fr,unsigned char *buf, int siz
 
 static off_t generic_tell(mpg123_handle *fr)
 {
+#ifndef NO_FEEDER
 	if(fr->rdat.flags & READER_BUFFERED)
 	fr->rdat.filepos = fr->rdat.buffer.fileoff+fr->rdat.buffer.pos;
+#endif
 
 	return fr->rdat.filepos;
 }
@@ -404,13 +411,20 @@ static off_t generic_tell(mpg123_handle *fr)
 static void stream_rewind(mpg123_handle *fr)
 {
 	if(fr->rdat.flags & READER_SEEKABLE)
-	fr->rdat.buffer.fileoff = fr->rdat.filepos = stream_lseek(fr,0,SEEK_SET);
+	{
+		fr->rdat.filepos = stream_lseek(fr,0,SEEK_SET);
+#ifndef NO_FEEDER
+		fr->rdat.buffer.fileoff = fr->rdat.filepos;
+#endif
+	}
+#ifndef NO_FEEDER
 	if(fr->rdat.flags & READER_BUFFERED)
 	{
 		fr->rdat.buffer.pos      = 0;
 		fr->rdat.buffer.firstpos = 0;
 		fr->rdat.filepos = fr->rdat.buffer.fileoff;
 	}
+#endif
 }
 
 /*
@@ -437,12 +451,136 @@ static off_t get_fileinfo(mpg123_handle *fr)
 	return len;
 }
 
-/* Let's work in nice 4K blocks, that may be nicely reusable (by malloc(), even). */
-#define BUFFBLOCK 4096
-
 #ifndef NO_FEEDER
 /* Methods for the buffer chain, mainly used for feed reader, but not just that. */
 
+
+static struct buffy* buffy_new(size_t size, size_t minsize)
+{
+	struct buffy *newbuf;
+	newbuf = malloc(sizeof(struct buffy));
+	if(newbuf == NULL) return NULL;
+
+	newbuf->realsize = size > minsize ? size : minsize;
+	newbuf->data = malloc(newbuf->realsize);
+	if(newbuf->data == NULL)
+	{
+		free(newbuf);
+		return NULL;
+	}
+	newbuf->size = 0;
+	newbuf->next = NULL;
+	return newbuf;
+}
+
+static void buffy_del(struct buffy* buf)
+{
+	if(buf)
+	{
+		free(buf->data);
+		free(buf);
+	}
+}
+
+/* Delete this buffy and all following buffies. */
+static void buffy_del_chain(struct buffy* buf)
+{
+	while(buf)
+	{
+		struct buffy* next = buf->next;
+		buffy_del(buf);
+		buf = next;
+	}
+}
+
+void bc_prepare(struct bufferchain *bc, size_t pool_size, size_t bufblock)
+{
+	bc_poolsize(bc, pool_size, bufblock);
+	bc->pool = NULL;
+	bc->pool_fill = 0;
+	bc_init(bc); /* Ensure that members are zeroed for read-only use. */
+}
+
+size_t bc_fill(struct bufferchain *bc)
+{
+	return (size_t)(bc->size - bc->pos);
+}
+
+void bc_poolsize(struct bufferchain *bc, size_t pool_size, size_t bufblock)
+{
+	bc->pool_size = pool_size;
+	bc->bufblock = bufblock;
+}
+
+void bc_cleanup(struct bufferchain *bc)
+{
+	buffy_del_chain(bc->pool);
+	bc->pool = NULL;
+	bc->pool_fill = 0;
+}
+
+/* Fetch a buffer from the pool (if possible) or create one. */
+static struct buffy* bc_alloc(struct bufferchain *bc, size_t size)
+{
+	/* Easy route: Just try the first available buffer.
+	   Size does not matter, it's only a hint for creation of new buffers. */
+	if(bc->pool)
+	{
+		struct buffy *buf = bc->pool;
+		bc->pool = buf->next;
+		buf->next = NULL; /* That shall be set to a sensible value later. */
+		buf->size = 0;
+		--bc->pool_fill;
+		debug2("bc_alloc: picked %p from pool (fill now %"SIZE_P")", buf, (size_p)bc->pool_fill);
+		return buf;
+	}
+	else return buffy_new(size, bc->bufblock);
+}
+
+/* Either stuff the buffer back into the pool or free it for good. */
+static void bc_free(struct bufferchain *bc, struct buffy* buf)
+{
+	if(!buf) return;
+
+	if(bc->pool_fill < bc->pool_size)
+	{
+		buf->next = bc->pool;
+		bc->pool = buf;
+		++bc->pool_fill;
+	}
+	else buffy_del(buf);
+}
+
+/* Make the buffer count in the pool match the pool size. */
+static int bc_fill_pool(struct bufferchain *bc)
+{
+	/* Remove superfluous ones. */
+	while(bc->pool_fill > bc->pool_size)
+	{
+		/* Lazyness: Just work on the front. */
+		struct buffy* buf = bc->pool;
+		bc->pool = buf->next;
+		buffy_del(buf);
+		--bc->pool_fill;
+	}
+
+	/* Add missing ones. */
+	while(bc->pool_fill < bc->pool_size)
+	{
+		/* Again, just work on the front. */
+		struct buffy* buf;
+		buf = buffy_new(0, bc->bufblock); /* Use default block size. */
+		if(!buf) return -1;
+
+		buf->next = bc->pool;
+		bc->pool = buf;
+		++bc->pool_fill;
+	}
+
+	return 0;
+}
+
+
 static void bc_init(struct bufferchain *bc)
 {
 	bc->first = NULL;
@@ -455,15 +593,14 @@ static void bc_init(struct bufferchain *bc)
 
 static void bc_reset(struct bufferchain *bc)
 {
-	/* free the buffer chain */
-	struct buffy *b = bc->first;
-	while(b != NULL)
+	/* Free current chain, possibly stuffing back into the pool. */
+	while(bc->first)
 	{
-		struct buffy *n = b->next;
-		free(b->data);
-		free(b);
-		b = n;
+		struct buffy* buf = bc->first;
+		bc->first = buf->next;
+		bc_free(bc, buf);
 	}
+	bc_fill_pool(bc); /* Ignoring an error here... */
 	bc_init(bc);
 }
 
@@ -473,53 +610,17 @@ static int bc_append(struct bufferchain *bc, ssize_t size)
 	struct buffy *newbuf;
 	if(size < 1) return -1;
 
-	newbuf = malloc(sizeof(struct buffy));
+	newbuf = bc_alloc(bc, size);
 	if(newbuf == NULL) return -2;
 
-	newbuf->realsize = size > BUFFBLOCK ? size : BUFFBLOCK;
-	newbuf->data = malloc(newbuf->realsize);
-	if(newbuf->data == NULL)
-	{
-		free(newbuf);
-		return -3;
-	}
-	newbuf->size = size;
-	newbuf->next = NULL;
 	if(bc->last != NULL)  bc->last->next = newbuf;
 	else if(bc->first == NULL) bc->first = newbuf;
 
 	bc->last  = newbuf;
-	bc->size += size;
+	debug3("bc_append: new last buffer %p with %"SSIZE_P" B (really %"SSIZE_P")", bc->last, (ssize_p)bc->last->size, (ssize_p)bc->last->realsize);
 	return 0;
 }
 
-#if 0
-/* Drop the last one (again).
-   This is not optimal but should happen on error situations only, anyway. */
-static void bc_drop(struct bufferchain *bc)
-{
-	struct buffy *cur = bc->first;
-	if(bc->first == NULL || bc->last == NULL) return;
-	/* Special case: only one buffer there. */
-	if(cur->next == NULL)
-	{
-		free(cur->data);
-		free(cur);
-		bc->first = bc->last = NULL;
-		bc->size  = 0;
-		return;
-	}
-	/* Find the pre-last buffy. If chain is consistent, this _will_ succeed. */
-	while(cur->next != bc->last){ cur = cur->next; }
-
-	bc->size -= bc->last->size;
-	free(bc->last->data);
-	free(bc->last);
-	cur->next = NULL;
-	bc->last  = cur;
-}
-#endif
-
 /* Append a new buffer and copy content to it. */
 static int bc_add(struct bufferchain *bc, const unsigned char *data, ssize_t size)
 {
@@ -528,23 +629,27 @@ static int bc_add(struct bufferchain *bc, const unsigned char *data, ssize_t siz
 	debug2("bc_add: adding %"SSIZE_P" bytes at %"OFF_P, (ssize_p)size, (off_p)(bc->fileoff+bc->size));
 	if(size >=4) debug4("first bytes: %02x %02x %02x %02x", data[0], data[1], data[2], data[3]);
 
-	/* Try to fill up the last buffer block. */
-	if(bc->last != NULL && bc->last->size < bc->last->realsize)
+	while(size > 0)
 	{
-		part = bc->last->realsize - bc->last->size;
-		if(part > size) part = size;
+		/* Try to fill up the last buffer block. */
+		if(bc->last != NULL && bc->last->size < bc->last->realsize)
+		{
+			part = bc->last->realsize - bc->last->size;
+			if(part > size) part = size;
 
-		memcpy(bc->last->data+bc->last->size, data, part);
-		bc->last->size += part;
-		size -= part;
-		bc->size += part;
+			debug2("bc_add: adding %"SSIZE_P" B to existing block %p", (ssize_p)part, bc->last);
+			memcpy(bc->last->data+bc->last->size, data, part);
+			bc->last->size += part;
+			size -= part;
+			bc->size += part;
+			data += part;
+		}
+
+		/* If there is still data left, put it into a new buffer block. */
+		if(size > 0 && (ret = bc_append(bc, size)) != 0)
+		break;
 	}
 
-
-	/* If there is still data left, put it into a new buffer block. */
-	if(size > 0 && (ret = bc_append(bc, size)) == 0)
-	memcpy(bc->last->data, data+part, size);
-
 	return ret;
 }
 
@@ -632,8 +737,7 @@ static void bc_forget(struct bufferchain *bc)
 
 		debug5("bc_forget: forgot %p with %lu, pos=%li, size=%li, fileoff=%li", (void*)b->data, (long)b->size, (long)bc->pos,  (long)bc->size, (long)bc->fileoff);
 
-		free(b->data);
-		free(b);
+		bc_free(bc, b);
 		b = n;
 	}
 	bc->first = b;
@@ -645,6 +749,7 @@ static void bc_forget(struct bufferchain *bc)
 static int feed_init(mpg123_handle *fr)
 {
 	bc_init(&fr->rdat.buffer);
+	bc_fill_pool(&fr->rdat.buffer);
 	fr->rdat.filelen = 0;
 	fr->rdat.filepos = 0;
 	fr->rdat.flags |= READER_BUFFERED;
@@ -728,12 +833,12 @@ static ssize_t buffered_fullread(mpg123_handle *fr, unsigned char *out, ssize_t
 	ssize_t gotcount;
 	if(bc->size - bc->pos < count)
 	{ /* Add more stuff to buffer. If hitting end of file, adjust count. */
-		unsigned char readbuf[BUFFBLOCK];
+		unsigned char readbuf[4096];
 		ssize_t need = count - (bc->size-bc->pos);
 		while(need>0)
 		{
 			int ret;
-			ssize_t got = fr->rdat.fullread(fr, readbuf, BUFFBLOCK);
+			ssize_t got = fr->rdat.fullread(fr, readbuf, sizeof(readbuf));
 			if(got < 0)
 			{
 				if(NOQUIET) error("buffer reading");
@@ -748,7 +853,7 @@ static ssize_t buffered_fullread(mpg123_handle *fr, unsigned char *out, ssize_t
 			}
 
 			need -= got; /* May underflow here... */
-			if(got < BUFFBLOCK) /* That naturally catches got == 0, too. */
+			if(got < sizeof(readbuf)) /* That naturally catches got == 0, too. */
 			{
 				if(VERBOSE3) fprintf(stderr, "Note: Input data end.\n");
 				break; /* End. */
@@ -939,8 +1044,10 @@ static int default_init(mpg123_handle *fr)
 
 	fr->rdat.read  = fr->rdat.r_read  != NULL ? fr->rdat.r_read  : posix_read;
 	fr->rdat.lseek = fr->rdat.r_lseek != NULL ? fr->rdat.r_lseek : posix_lseek;
+#ifndef NO_ICY
 	/* ICY streams of any sort shall not be seekable. */
 	if(fr->p.icy_interval > 0) fr->rdat.lseek = nix_lseek;
+#endif
 
 	fr->rdat.filelen = get_fileinfo(fr);
 	fr->rdat.filepos = 0;
@@ -994,12 +1101,16 @@ static int default_init(mpg123_handle *fr)
 
 void open_bad(mpg123_handle *mh)
 {
+	debug("open_bad");
 #ifndef NO_ICY
 	clear_icy(&mh->icy);
 #endif
 	mh->rd = &bad_reader;
 	mh->rdat.flags = 0;
+#ifndef NO_FEEDER
 	bc_init(&mh->rdat.buffer);
+#endif
+	mh->rdat.filelen = -1;
 }
 
 int open_feed(mpg123_handle *fr)
@@ -1022,6 +1133,8 @@ int open_feed(mpg123_handle *fr)
 	fr->rd = &readers[READER_FEED];
 	fr->rdat.flags = 0;
 	if(fr->rd->init(fr) < 0) return -1;
+
+	debug("feed reader init successful");
 	return 0;
 #endif /* NO_FEEDER */
 }
